from .base import QS, QsNode
import urllib.parse as up
import typing as t

LIST_FORMAT_OPTIONS = ['indices', 'brackets', 'repeat', 'comma']


class QsStringifier(QS):
    _af: str = 'indices'

    def __init__(self, depth: int = 5, parameter_limit: int = 1000, allow_dots: bool = False, allow_sparse: bool = False, array_limit: int = 20, parse_arrays: bool = False, allow_empty: bool = False, comma: bool = False, array_format: str = 'indices'):
        super().__init__(depth, parameter_limit, allow_dots,
                         array_limit, parse_arrays, allow_empty, comma)
        if array_format not in LIST_FORMAT_OPTIONS:
            raise ValueError(
                f'array_format must be one of {LIST_FORMAT_OPTIONS}')
        self._af = array_format

    def stringify(self, data: dict, filter: list = None) -> list[tuple]:
        """
        Process a dictionary into a query string

        Args:
            data (dict): dictionary to process
            filter (list): list of keys to filter

        Returns:    
            str: query string
        """
        _q_args = {}
        for item in data:
            node = QsNode.load(
                item,
                data[item],
                filter=filter)
            if node is not None:
                self._qs_tree[node.key] = node
        for item in self._qs_tree.values():
            for arg in self._get_arg(item):
                if arg is None:
                    continue
                prepared_arg = self._transform_arg(arg)
                if prepared_arg[0] in _q_args:
                    _q_args[prepared_arg[0]] = [_q_args[prepared_arg[0]], prepared_arg[1]] if isinstance(
                        _q_args[prepared_arg[0]], str) else [*_q_args[prepared_arg[0]], prepared_arg[1]]
                else:
                    _q_args[prepared_arg[0]] = prepared_arg[1]
        res = []
        for key, value in _q_args.items():
            if isinstance(value, list) and (self._af == 'repeat' or self._af == 'brackets'):
                res += [(key, v) for v in value]
            else:
                res.append((key, str(value)))
        return res

    def _format_key(self, key: list) -> str | None:
        """
        Format key based on array_format into proper string representation recursively

        Args:
            key (list): list of keys

        Returns:
            str: formatted key
        """
        if len(key) == 1:
            # decide based on array_format
            k = key[0]
            k = key[0]
            if isinstance(k, int) and (self._af == 'repeat' or self._af == 'brackets'):
                return '[]' if self._af == 'brackets' else None
            else:
                return f'[{k}]' if not self._allow_dots else k
        else:
            current_key = key.pop()
            formatted = self._format_key(key)
            if formatted is None:
                return f'[{current_key}]' if not self._allow_dots else f'.{current_key}'
            if self._allow_dots:
                return f'{current_key}.{formatted}'
            return f'[{current_key}]{formatted}'

    def _get_arg(self, node: QsNode) -> tuple[str, t.Any]:
        """
        Get argument from node recursively as generator

        Args:
            node (QsNode): node to process

        Returns:
            tuple[str, Any]: argument
        """
        if node.is_leaf():
            if node.is_empty():
                yield None
            else:
                yield ([node.key], node.value)
        elif node.is_default_array() and self._af == 'comma':
            yield ([node.key], ','.join([str(child.value) for child in node.children]))
        else:
            for child in node.children:
                for arg in self._get_arg(child):
                    yield arg if arg is None else ([*arg[0], node.key], arg[1])

    def _transform_arg(self, arg: tuple[list, t.Any]) -> tuple[str, t.Any]:
        """
        Transform argument into proper string format

        Args:
            arg (tuple[list, Any]): argument

        Returns:
            tuple[str, Any]: transformed argument
        """
        notation, value = arg
        if len(notation) == 1:
            return (notation[0], value)
        key = notation.pop()
        nesting = self._format_key(notation)
        if self._allow_dots:
            key = f'{key}{("." + nesting) if nesting is not None else ""}'
        else:
            key = f'{key}{nesting or ""}'
        return (key, value)


def stringify(
        data: dict,
        allow_dots: bool = False,
        encode: bool = True,
        delimiter: str = '&',
        encode_values_only: bool = False,
        array_format: str = 'indices',
        sort: bool = False,
        sort_reverse: bool = False,
        charset: str = 'utf-8',
        filter: list = None,
        charset_sentinel: bool = False,
) -> str:
    """
    Process a dictionary into a query string

    Args:
        data (dict): dictionary to process
        allow_dots (bool): use dot notation instead of brackets
        encode (bool): encode values and keys
        delimiter (str): delimiter to use
        encode_values_only (bool): encode only values
        array_format (str): array format to use
        sort (bool): sort query string
        sort_reverse (bool): sort query string in reverse
        charset (str): charset to use
        filter (list): list of keys to filter
        charset_sentinel (bool): add utf8 sentinel

    Returns:    
        str: query string
    """
    qs = QsStringifier(
        allow_dots=allow_dots,
        array_format=array_format
    )
    _qt = qs.stringify(data, filter=filter)
    if charset_sentinel:
        _qt.insert(0, ('utf8', '✓'))
    if sort:
        _qt = sorted(_qt, key=lambda x: x[0], reverse=sort_reverse)
    if encode:
        return delimiter.join([qs._q(k, v, charset=charset) for k, v in _qt])
    elif encode_values_only:
        return delimiter.join([f'{k}={up.quote(v, encoding=charset, errors="xmlcharrefreplace")}' for k, v in _qt])
    return delimiter.join([f'{k}={v}' for k, v in _qt])



import typing as t
import urllib.parse as up
import re
from .base import QS, Unparsable, UnbalancedBrackets, QsNode, EmptyKey, ArrayLimitReached
from html import unescape as unescape_html


t_Delimiter = t.Union[str, t.Pattern[str]]


class ArrayParse:
    """
    Parses array notation into a tree like structure.
    """
    _depth: int = 5
    _limit: int = 20

    @classmethod
    def process(cls, notation: list, val: str, depth: int = 5, max_limit: int = 20) -> QsNode:
        """
        Parses array notation for single key-value pair.

        Args:
            notation (list): list of keys (in nested order)
            val (str): value to assign to the last key
            depth (int): max depth of nested objects
            max_limit (int): max number of elements in array

        Returns:
            list: parsed item into dictionary
        """
        cls._limit = max_limit
        cls._depth = depth
        return cls.process_notation(notation, val)

    @classmethod
    def process_notation(cls, notation: list[str], val: t.Any) -> QsNode:
        """
        Parses array notation recursively.

        Args:
            notation (list): list of remaining keys (in nested order)
            val (str): value to assign to the last key

        Returns:
            QsNode: tree like structure of parsed data

        Raises:
            ArrayLimitReached: if array limit is reached
        """
        if not notation:
            return val
        current = notation[0]
        if current.isdigit():
            if int(current) > cls._limit:
                raise ArrayLimitReached('Array limit reached')
            return QsNode(int(current), cls.process_notation(notation[1:], val))
        elif current == '':
            res = cls.process_notation(notation[1:], val)
            if not isinstance(res, QsNode):
                return QsNode(None, res)
            if not res.has_int_key() and res.key is not None:
                # initialization of array without index - set index None to be handled by parser
                return QsNode(None, res)
            return res
        else:
            return QsNode(current, cls.process_notation(notation[1:], val))


class LHSParse:
    """
    Parses left hand side notation into a tree like structure.
    """

    _depth: int = 5
    _dots: bool = False
    _allow_empty: bool = False

    @classmethod
    def process(
        cls,
        notation: list,
        val: str,
        depth: int = 5,
        allow_empty: bool = False,
        allow_dots: bool = False
    ) -> QsNode:
        """
        Parses left hand side notation for single key-value pair.

        Args:
            notation (list): list of keys (in nested order)
            val (str): value to assign to the last key
            depth (int): max depth of nested objects
            allow_empty (bool): allow empty keys
            allow_dots (bool): allow dot notation
        """
        cls._depth = depth
        cls._allow_empty = allow_empty
        cls._dots = allow_dots
        return cls.process_notation(notation, val)

    @classmethod
    def process_notation(cls, notation: list[str], val: t.Any) -> QsNode:
        """
        Parses left hand side notation recursively.

        Args:
            notation (list): list of remaining keys (in nested order)
            val (str): value to assign to the last key

        Returns:
            dict: parsed sub-dictionary

        Raises:
            EmptyKey: if empty key is found and not allowed
        """
        if not notation:
            return val
        current = notation[0]
        if cls._depth < 0:
            # join current and rest of notation
            current_key = cls._max_depth_key(notation)
            return QsNode(current_key, val)
        cls._depth -= 1
        return QsNode(current, cls.process_notation(notation[1:], val))

    @classmethod
    def _max_depth_key(cls, notation: list[str]) -> str:
        """
        Returns a key for max depth reached. 

        Args:
            notation (list): list of remaining keys (in nested order)

        Returns:
            str: key for max depth reached
        """
        if len(notation) == 1:
            return notation[0]
        if cls._dots:
            return f'{".".join(notation)}'
        return f'[{"][".join(notation)}]'


class QsParser(QS):
    _parse_primitive: bool = False
    _primitive_strict: bool = True

    def __init__(
            self,
            depth: int = 5,
            parameter_limit: int = 1000,
            allow_dots: bool = False,
            array_limit: int = 20,
            parse_arrays: bool = False,
            allow_empty: bool = False,
            comma: bool = False,
            parse_primitive: bool = False,
            primitive_strict: bool = True
    ):
        super().__init__(depth, parameter_limit, allow_dots,
                         array_limit, parse_arrays, allow_empty, comma)
        self._parse_primitive = parse_primitive
        self._primitive_strict = primitive_strict

    def parse(self, args: list[tuple]):
        """
        Parses a list of key-value pairs into a dictionary of nested tree-like structures.

        Args:
            args (list): list of key-value pairs
        """
        for arg in args:
            parse_func = self._parse_array if self._parse_arrays else self._parse_lhs
            k, v = arg
            if self._comma:
                v = re.split(',', v)
                v = str(v[0]) if len(v) == 1 else v
            parse_func(k, v)

    @property
    def args(self) -> dict[str, str]:
        """
        Returns nested dictionary as representation of argument tree.
        """
        return {
            k.key: k.serialize() for k in self._qs_tree.values()
        }

    @staticmethod
    def _find_charset_sentinel(args: list[str]) -> str | None:
        """
        Finds charset sentinel argument. If found, it is removed from original list.

        Args:
            args (list): list of arguments

        Returns:
            str: charset sentinel or None if not found
        """
        utf_idx = None
        for idx, arg in enumerate(args):
            if arg.split('=')[0] == 'utf8':
                utf_idx = idx
                break
        if utf_idx is None:
            return None
        val = args.pop(utf_idx).split('=')[1]
        if up.unquote(val, encoding='utf-8') == '✓':
            return 'utf-8'
        elif unescape_html(up.unquote(val, encoding='iso-8859-1')) == '✓':
            return 'iso-8859-1'
        else:
            raise Unparsable('Unable to parse charset sentinel')

    @staticmethod
    def _from_array_like(v: str | list) -> list | str:
        """
        Converts array-like string to list.

        Args:
            v (str): string to convert

        Returns:
            list: converted list
        """

        if isinstance(v, str) and (v.startswith('[') and v.endswith(']')):
            v = v.rstrip(']').lstrip('[')
            return v.split(',')
        return v

    @staticmethod
    def _check_brackets(k: str) -> None:
        """
        Checks if brackets are balanced.

        Args:
            k (str): key to check

        Raises:
            UnbalancedBrackets: if brackets are unbalanced
            Unparsable: if nesting notation is broken
        """
        brackets = [char for char in k if char in '[]']
        # check if brackets are balanced
        bracket_count = 0
        for bracket in brackets:
            if bracket == '[' and bracket_count > 0:
                raise UnbalancedBrackets(
                    'Using brackets as key is not allowed')
            if bracket == ']' and bracket_count == 0:
                raise UnbalancedBrackets('Unbalanced brackets')
            bracket_count += 1 if bracket == '[' else -1
        if bracket_count != 0:
            raise UnbalancedBrackets('Unbalanced brackets')
        if brackets and not k.endswith(']'):
            raise Unparsable('Nesting notation broken')

    def _parse_array(self, k: str, v: str | list) -> None:
        """
        Parses key with array notation into a list of nested keys.

        Args:
            k (str): key to parse
            v (str): value to assign to the key
        """
        v = self._process_primitive(v)
        notation = self._split_key(k)
        if notation is None:
            # continue as default lhs
            self._parse_arrays = False
            self._to_obj()
            return self._parse_lhs(k, v)
        if re.match(r'[^\d]+', notation[1]):
            self._parse_arrays = False
            self._to_obj()
            return self._parse_lhs(k, v)
        try:
            item = ArrayParse.process(notation, v)
            item.set_index(self._qs_tree.get(item.key, None),
                           array_limit=self._array_limit)
        except ArrayLimitReached:
            self._parse_arrays = False
            self._to_obj()
            return self._parse_lhs(k, v)
        if item.key not in self._qs_tree:
            if len(self._qs_tree) >= self._parameter_limit:
                return
            self._qs_tree[item.key] = item
        else:
            self._qs_tree[item.key].update(item)

    def _parse_lhs(self, k: str, v: str | list) -> None:
        """
        Parses key with brackets notation into a list of nested keys.

        Args:
            k (str): key to parse
            v (str): value to assign to the key
        """
        v = self._process_primitive(v)
        notation = self._split_key(k)
        if notation is None:
            raise Unparsable('Unable to parse key')
        data = LHSParse.process(
            notation, v, depth=self._max_depth, allow_empty=self._allow_empty, allow_dots=self._allow_dots)
        if data.key not in self._qs_tree:
            if len(self._qs_tree) >= self._parameter_limit:
                return
            self._qs_tree[data.key] = data
        else:
            self._qs_tree[data.key].update(data)

    def _process_primitive(self, v: str | list) -> t.Any:
        """
        Processes primitive value into a proper type.

        Args:
            v (str): value to process

        Returns:
            Any: processed value
        """
        v = self._from_array_like(v)
        if not self._parse_primitive:
            return v
        if isinstance(v, list):
            return [self._process_primitive(item) for item in v]
        if v.isdigit():
            return int(v)
        if not self._primitive_strict:
            if v.lower() in ['true', 'false']:
                return v.lower() == 'true'
            if v.lower() in ['null', 'none']:
                return None
        else:
            if v in ['true', 'false']:
                return v == 'true'
            if v in ['null', 'None']:
                return None
        try:
            return float(v)
        except ValueError:
            return v

    def _split_key(self, k: str) -> list[str] | None:
        """
        Splits key into a main key and a list of nested keys.

        Args:
            k (str): key to split

        Returns:
            tuple[str, list[str]]: main key and list of nested keys or None, None if key is not parsable with current settings
        """
        if self._parse_arrays:
            QsParser._check_brackets(k)
            match_pattern = r'(\w+)(\[(.*)\])+' if not self._allow_empty else r'(\w*)(\[(.*)\])+'
            match = re.match(match_pattern, k)
            notation = re.findall(r'\[(.*?)\]', k)
            return [match.group(1)] + notation if match else None
        if self._allow_dots:
            match = re.match(
                r'(\w+)(\.\w+)+', k) if not self._allow_empty else re.match(r'(\w*)(\.(\w*))', k)
            notation = re.findall(
                r'\.(\w+)', k) if not self._allow_empty else re.findall(r'\.(\w*)', k)
            if match:
                return [match.group(1)] + notation
        QsParser._check_brackets(k)
        match_pattern = r'^(\w+)(\[\w+\])*$' if not self._allow_empty else r'^(\w*)(\[\w*\])*$'
        match = re.match(match_pattern, k)
        notation = re.findall(
            r'\[(\w+)\]', k) if not self._allow_empty else re.findall(r'\[(\w*)\]', k)
        return [match.group(1)] + notation if match else None

    def _to_obj(self) -> None:
        """
        Transforms array-like dictionary into object-like dictionary, recursively.

        Args:
            arg (dict): dictionary to transform

        Returns:    
            dict: transformed dictionary
        """
        for v in self._qs_tree.values():
            v.to_object_notation()


def unpack_payload(payload: str | bytes) -> str:
    """
    Unpacks payload from bytes to string.

    Args:
        payload (str): payload to unpack

    Returns:
        str: unpacked payload
    """
    if isinstance(payload, bytes):
        return payload.decode('utf-8')
    return payload


def parse_from_dict(
    data: dict[str, str],
    **kw,
):
    """
    Parses a filter from dictionary (args).

    Args:
        data (dict): dictionary to parse
        **kw: keyword arguments - for parser class
    """
    try:
        parser = QsParser(**kw)
        parser.parse(data.items())
        return parser.args
    except (Unparsable, UnbalancedBrackets):
        return up.parse_qs(
            up.urlencode(data),
            keep_blank_values=kw.get('allow_empty', False),
            max_num_fields=kw.get('parameter_limit', 1000),
            separator=kw.get('delimiter', '&')
        )


def parse(
        data: str | bytes,
        from_url: bool = False,
        delimiter: t_Delimiter = '&',
        depth: int = 5,
        parameter_limit: int = 1000,
        allow_dots: bool = False,
        array_limit: int = 20,
        parse_arrays: bool = False,
        allow_empty: bool = False,
        charset: str = 'utf-8',
        charset_sentinel: bool = False,
        interpret_numeric_entities: bool = False,
        parse_primitive: bool = False,
        primitive_strict: bool = True,
        comma: bool = False):
    """
    Parses a string into a dictionary.

    Args:
        data (str): string to parse
        from_url (bool): if True, data is parsed from url
        delimiter (str): delimiter to use
        depth (int): max depth of nested objects
        parameter_limit (int): max number of parameters
        allow_dots (bool): allow dot notation
        array_limit (int): max number of elements in array
        parse_arrays (bool): parse arrays
        allow_empty (bool): allow empty keys and values
        charset (str): charset to use
        charset_sentinel (bool): if True, charset is parsed from data if charset sentinel is found
        interpret_numeric_entities (bool): if True, numeric entities are interpreted into unicode characters
        parse_primitive (bool): if True, primitive values are parsed into proper types
        primitive_strict (bool): if True, primitive values are parsed into proper types strictly (affects bool and nonetype)
        comma (bool): if True, comma is used as delimiter

    Returns:
        dict: parsed data
    """
    data = unpack_payload(data)
    if from_url:
        qs = up.urlparse(data).query
    else:
        qs = data
    args = []
    try:
        query_args = re.split(delimiter, qs)
        if charset_sentinel:
            charset = QsParser._find_charset_sentinel(query_args) or charset
        for arg in query_args:
            args.append(QsParser._unq(
                arg, charset, interpret_numeric_entities))
        parser = QsParser(depth, parameter_limit, allow_dots,
                          array_limit, parse_arrays, allow_empty, comma, parse_primitive, primitive_strict)
        parser.parse(args)
        return parser.args
    except (Unparsable, UnbalancedBrackets):
        return up.parse_qs(
            qs,
            keep_blank_values=allow_empty,
            max_num_fields=parameter_limit,
            separator=delimiter)



import unittest
import sys
sys.path.append(".")


class StringifyTest(unittest.TestCase):

    def test_basic_stringify_objects(self):
        import src.qstion as qs

        self.assertEqual(
            qs.stringify({'a': 'b'}),
            'a=b')

        # basic nested
        self.assertEqual(
            qs.stringify({'a': {'b': 'c'}}),
            'a%5Bb%5D=c')

        # encoding option
        self.assertEqual(
            qs.stringify({'a': {'b': 'c'}}, encode=False),
            'a[b]=c')

        # encoding values only
        self.assertEqual(
            qs.stringify(
                {'a': 'b', 'c': ['d', 'e=f'], 'f': [['g'], ['h']]},
                encode_values_only=True, encode=False
            ),
            'a=b&c[0]=d&c[1]=e%3Df&f[0][0]=g&f[1][0]=h'
        )

        # When arrays are stringified, by default they are given explicit indices:
        self.assertEqual(
            qs.stringify({'a': ['b', 'c', 'd']}, encode=False),
            'a[0]=b&a[1]=c&a[2]=d')

        # Options for indices is not available, use array_format option instead:
        self.assertEqual(
            qs.stringify({'a': ['b', 'c', 'd']},
                         array_format='indices', encode=False),
            'a[0]=b&a[1]=c&a[2]=d')

        self.assertEqual(
            qs.stringify({'a': ['b', 'c', 'd']},
                         array_format='brackets', encode=False),
            'a[]=b&a[]=c&a[]=d')

        self.assertEqual(
            qs.stringify({'a': ['b', 'c', 'd']},
                         array_format='repeat', encode=False),
            'a=b&a=c&a=d')

        self.assertEqual(
            qs.stringify({'a': ['b', 'c', 'd']},
                         array_format='comma', encode=False),
            'a=b,c,d')

        # When objects are stringified, by default they use bracket notation:
        self.assertEqual(
            qs.stringify({'a': {'b': {'c': 'd', 'e': 'f'}}}, encode=False),
            'a[b][c]=d&a[b][e]=f')

        # You may override this to use dot notation by setting the allowDots option to true:
        self.assertEqual(
            qs.stringify({'a': {'b': {'c': 'd', 'e': 'f'}}},
                         allow_dots=True, encode=False),
            'a.b.c=d&a.b.e=f')

        # Empty strings and null values will omit the value, but the equals sign (=) remains in place:
        self.assertEqual(
            qs.stringify({'a': ''}, encode=False), 'a=')

    def test_advanced_stringify_objects(self):
        import src.qstion as qs

        # Key with no values (such as an empty object or array) will return nothing:

        self.assertEqual(
            qs.stringify({'a': []}, encode=False), '')

        self.assertEqual(
            qs.stringify({'a': {}}, encode=False), '')

        self.assertEqual(
            qs.stringify({'a': [{}]}, encode=False), '')

        self.assertEqual(
            qs.stringify({'a': {'b': {}}}, encode=False), '')

        self.assertEqual(
            qs.stringify({'a': {'b': []}}, encode=False), '')

        # Properties that are set to `undefined` will be omitted entirely
        # NOTE undefined is not available in python -> handled as null -> None

        self.assertEqual(
            qs.stringify({'a': None}, encode=False), '')

        # The delimiter may be overridden with stringify as well:

        self.assertEqual(
            qs.stringify({'a': 'b', 'c': 'd'}, delimiter=';'),
            'a=b;c=d')

        # You may use the sort option to affect the order of parameter keys -> only functions as bool option:

        self.assertEqual(
            qs.stringify({'a': 'c', 'z': 'y', 'b': 'f'}, sort=True),
            'a=c&b=f&z=y')

        # also implemented reverse sort

        self.assertEqual(
            qs.stringify({'a': 'c', 'z': 'y', 'b': 'f'},
                         sort=True, sort_reverse=True),
            'z=y&b=f&a=c')

        # NOTE passing functions as filter is not available, use list as value
        self.assertEqual(
            qs.stringify({'a': 'b', 'c': 'd', 'e': 'f'},
                         filter=['a', 'e'], encode=False),
            'a=b&e=f')

        self.assertEqual(
            qs.stringify({'a': ['b', 'c', 'd'], 'e': 'f'},
                         filter=['a', 0, 2], encode=False),
            'a[0]=b&a[2]=d')

        # If you're communicating with legacy systems, you can switch to iso-8859-1 using the charset option:
        self.assertEqual(
            qs.stringify({'æ': 'æ'}, charset='iso-8859-1'),
            '%E6=%E6')

        # Characters that don't exist in iso-8859-1 will be converted to numeric entities, similar to what browsers do:

        self.assertEqual(
            qs.stringify({'a': '☺'}, charset='iso-8859-1'),
            'a=%26%239786%3B')

        # You can use the charsetSentinel option to announce the character by including an utf8=✓ parameter with the proper encoding if the checkmark

        self.assertEqual(
            qs.stringify({'a': '☺'}, charset_sentinel=True),
            'utf8=%E2%9C%93&a=%E2%98%BA')

        self.assertEqual(
            qs.stringify({'a': 'æ'}, charset='iso-8859-1',
                         charset_sentinel=True),
            'utf8=%26%2310003%3B&a=%E6'
        )

        # Dealing with special character sets : visit https://docs.python.org/3/library/codecs.html#standard-encodings

        # array format as brackets for array with shifted index
        self.assertEqual(
            qs.stringify({'a': {1: 'b', 2: 'c'}},
                         array_format='brackets', encode=False),
            'a[]=b&a[]=c'
        )

        # use comma as delimiter
        self.assertEqual(
            qs.stringify({'a': {1: 'b', 2: 'c'}},
                         array_format='comma', encode=False),
            'a=b,c'
        )

        # test unkown options for array_format
        self.assertRaises(ValueError, qs.stringify,
                          {'a': {1: 'b', 2: 'c'}}, array_format='unknown')

        # test bracket notation on basic object
        self.assertEqual(
            qs.stringify({'a': {'b': 'c'}},
                         array_format='brackets', encode=False),
            'a[b]=c'
        )

        # test bracket notation on deeper nesting
        self.assertEqual(
            qs.stringify({'a': {'b': {'c': 'd'}}},
                         array_format='brackets', encode=False),
            'a[b][c]=d'
        )

        # use comma delimiter in nested objects
        self.assertEqual(
            qs.stringify({'a': {'b': ['c', 'd']}},
                         array_format='comma', encode=False),
            'a[b]=c,d'
        )

        # use repeat on nested objects
        self.assertEqual(
            qs.stringify({'a': {'b': {1: 'c',2: 'd'}}}, array_format='repeat', encode=False),
            'a[b]=c&a[b]=d'
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import sys
sys.path.append(".")


class ParserTest(unittest.TestCase):

    def test_basic_parse_objects(self):
        import src.qstion as qs
        # tests based on README of qs in npm: https://www.npmjs.com/package/qs?activeTab=readme
        obj = qs.parse('a=c')
        self.assertDictEqual(obj, {'a': 'c'})

        self.assertDictEqual(qs.parse('foo[bar]=baz'), {'foo': {'bar': 'baz'}})

        # plain objects not supported -> unknown data type

        # uri encoded strings
        self.assertDictEqual(qs.parse('a%5Bb%5D=c'), {'a': {'b': 'c'}})

        # double nested
        self.assertDictEqual(
            qs.parse('foo[bar][baz]=foobarbaz'),
            {'foo': {'bar': {'baz': 'foobarbaz'}}})

        # default depth nesting -> 5
        self.assertDictEqual(
            qs.parse('a[b][c][d][e][f][g][h][i]=j'),
            {'a': {'b': {'c': {'d': {'e': {'f': {'[g][h][i]': 'j'}}}}}}})

        # overriding depth nesting -> 1
        self.assertDictEqual(
            qs.parse('a[b][c][d][e][f][g][h][i]=j', depth=1),
            {'a': {'b': {'[c][d][e][f][g][h][i]': 'j'}}})

        # parameter limited
        self.assertDictEqual(
            qs.parse('a=b&c=d', parameter_limit=1),
            {'a': 'b'})

        # parameter limited but multiple values
        self.assertDictEqual(
            qs.parse('a=b&a=c', parameter_limit=1),
            {'a': ['b', 'c']})

        # parameter limited but multiple values after reaching limit
        self.assertDictEqual(
            qs.parse('a=b&b=c&a=d', parameter_limit=1),
            {'a': ['b', 'd']})

        # optional delimiter
        self.assertDictEqual(
            qs.parse('a=b;c=d', delimiter=';'),
            {'a': 'b', 'c': 'd'})

        # regular expression delimiter
        self.assertDictEqual(
            qs.parse('a=b;c=d,e=f', delimiter=r'[;,]'),
            {'a': 'b', 'c': 'd', 'e': 'f'})

        # dot notation
        self.assertDictEqual(
            qs.parse('a.b=c', allow_dots=True),
            {'a': {'b': 'c'}})

        # using old charset
        self.assertDictEqual(
            qs.parse('a=%A7', charset='iso-8859-1'),
            {'a': '§'})

        # Some services add an initial utf8=✓ value to forms so that old Internet Explorer versions are more likely to submit the form as utf-8.
        # qs supports this mechanism via the charsetSentinel option.
        # If specified, the utf8 parameter will be omitted from the returned object.
        # It will be used to switch to iso-8859-1/utf-8 mode depending on how the checkmark is encoded.
        # Important: When you specify both the charset option and the charsetSentinel option,
        #  the charset will be overridden when the request contains a utf8 parameter from which the actual charset can be deduced.
        # In that sense the charset will behave as the default charset rather than the authoritative charset.\
        self.assertDictEqual(
            qs.parse('utf8=%E2%9C%93&a=%C3%B8',
                     charset_sentinel=True, charset='iso-8859-1'),
            {'a': 'ø'})

        self.assertDictEqual(
            qs.parse('utf8=%26%2310003%3B&a=%F8',
                     charset_sentinel=True, charset='utf-8'),
            {'a': 'ø'})

        # interpretNumericEntities &#...; syntax to the actual character
        self.assertDictEqual(
            qs.parse('a=%26%239786%3B',
                     interpret_numeric_entities=True, charset='iso-8859-1'),
            {'a': '☺'})

    def test_advanced_parse_objects(self):
        import src.qstion as qs

        # test parse from url
        self.assertDictEqual(
            qs.parse('http://example.com/?a=b', from_url=True),
            {'a': 'b'})

        # test unparsable qs
        self.assertDictEqual(
            qs.parse('this_is_unparsable'),
            {})

        # two separate nested objects
        self.assertDictEqual(
            qs.parse('a[b]=c&d[e]=f'),
            {'a': {'b': 'c'}, 'd': {'e': 'f'}})

        # common base key for nested objects
        self.assertDictEqual(
            qs.parse('a[b]=c&a[d]=e'),
            {'a': {'b': 'c', 'd': 'e'}})

        # common nested key for nested objects
        self.assertDictEqual(
            qs.parse('a[b][c]=d&a[b][e]=f'),
            {'a': {'b': {'c': 'd', 'e': 'f'}}})

        # non equal nesting after creating basic object
        self.assertDictEqual(
            qs.parse('a[b]=c&a[b][d]=e'),
            {'a': {'b': {'d': 'e', 'c': True}}})

        # non equal basic object after nesting
        self.assertDictEqual(
            qs.parse('a[b][c]=d&a[b]=e'),
            {'a': {'b': {'c': 'd', 'e': True}}})

        # multiple values for same nested key
        self.assertDictEqual(
            qs.parse('a[b]=c&a[b]=d'),
            {'a': {'b': ['c', 'd']}})

        # multiple values for same nested key with different nesting
        self.assertDictEqual(
            qs.parse('a[b]=c&a[b]=f&a[b][d]=e'),
            {'a': {'b': {str(['c', 'f']): True, 'd': 'e'}}})

        # three elements for same nested key
        self.assertDictEqual(
            qs.parse('a[b]=c&a[b]=f&a[b]=e'),
            {'a': {'b': ['c', 'f', 'e']}})

        # test array as string value
        self.assertDictEqual(
            qs.parse('a=[b,c]'),
            {'a': ['b', 'c']})

        # test very empty objects
        self.assertDictEqual(
            qs.parse('a[]=&b[]=', allow_empty=True),
            {'a': {'': ''}, 'b': {'': ''}})

        # test charset sentinel arg, but not as leading arg
        self.assertDictEqual(
            qs.parse('a=%A7&utf8=%26%2310003',
                     charset='utf-8', charset_sentinel=True),
            {'a': '§'}
        )

        # Special case : nesting to non-nested argument
        # NOTE: assuming that querystring 'a[b]=c&a[b][d]=e' parses into
        # {'a': {'b': {'d': 'e', 'c': True}}}
        # Handling of 'overloading' argument in such way: 'a[b]=c&a[b][c]=d'
        # will result in combining both True and 'd' into array
        self.assertDictEqual(
            qs.parse('a[b]=c&a[b][c]=d'),
            {'a': {'b': {'c': [True, 'd']}}})

        # test not allowing empty keys
        self.assertDictEqual(
            qs.parse('a[]=b&a[]=c&'),
            {'a[]': ['b', 'c']})

        # test dot notation with max depth
        self.assertDictEqual(
            qs.parse('a.b.c.d=e', allow_dots=True, depth=1),
            {'a': {'b': {'c.d': 'e'}}})

        # test exact depth
        self.assertDictEqual(
            qs.parse('a.b.c.d=e', allow_dots=True, depth=2),
            {'a': {'b': {'c': {'d': 'e'}}}})

        # test charset_sentinel option without utf8 present
        self.assertDictEqual(
            qs.parse('a=%A7', charset_sentinel=True, charset='iso-8859-1'),
            {'a': '§'})

        # test charset_sentinel option with utf8 present and '✓' encoded in unkown charset
        self.assertDictEqual(
            qs.parse('utf8=%2BJxM-&a=b', charset_sentinel=True),
            {'a': ['b'], 'utf8': ['+JxM-']}
        )

        # test unbalanced brackets - and thus unparsable by qstion
        self.assertDictEqual(
            qs.parse('a[b=c'),
            {'a[b': ['c']})

        # test balanced but nested brackets
        self.assertDictEqual(
            qs.parse('a[b[c]]=d'),
            {'a[b[c]]': ['d']})

        # test only closing brackets
        self.assertDictEqual(
            qs.parse('a[b]]]=d'),
            {'a[b]]]': ['d']})

        # test correct brackets but broken nested notation
        self.assertDictEqual(
            qs.parse('a[b][c]d=e'),
            {'a[b][c]d': ['e']})

    def test_basic_parse_arrays(self):
        import src.qstion as qs
        # NOTE
        # since python list would by default not support sparse indexes, dictionary is used instead
        # however, notation is still strictly checked so array notation is accepted
        # and array-like dictionaries should only contain integer keys (unless notation was broken and
        # notation is combined - with basic objects - in which case the notation is ignored)
        # basic
        self.assertDictEqual(
            qs.parse('a[]=b&a[]=c', parse_arrays=True),
            {'a': {0: 'b', 1: 'c'}})

        # with indexes
        self.assertDictEqual(
            qs.parse('a[1]=c&a[0]=b', parse_arrays=True),
            {'a': {0: 'b', 1: 'c'}})

        # with indexes in sparse order
        # NOTE that order is not important in python dictionary
        self.assertDictEqual(
            qs.parse('a[1]=b&a[15]=c', parse_arrays=True),
            {'a': {1: 'b', 15: 'c'}})

        # sparse arrays such as [,,a,b] are not supported

        # array with empty string
        self.assertDictEqual(
            qs.parse('a[]=&a[]=b', parse_arrays=True, allow_empty=True),
            {'a': {0: '', 1: 'b'}})

        # also with index
        self.assertDictEqual(
            qs.parse('a[0]=b&a[1]=&a[2]=c',
                     parse_arrays=True, allow_empty=True),
            {'a': {0: 'b', 1: '', 2: 'c'}})

        # overflowing max array length
        self.assertDictEqual(
            qs.parse('a[100]=b', parse_arrays=True),
            {'a': {'100': 'b'}})

        # overriding max array length
        self.assertDictEqual(
            qs.parse('a[1]=b', parse_arrays=True, array_limit=0),
            {'a': {'1': 'b'}})

        # NOTE: disabling parsing arrays and trying to parse array
        # notation is such case: a[]=b will not result in {'a': {'0': 'b'}}
        # but rather {'a': {'': 'b'}} -> only if empty string is allowed
        # else is considered unparsable and used urllib.parse_qs instead
        self.assertDictEqual(
            qs.parse('a[]=b', parse_arrays=False, allow_empty=True),
            {'a': {'': 'b'}})

        self.assertDictEqual(
            qs.parse('a[]=b', parse_arrays=False, allow_empty=False),
            {'a[]': ['b']})

        # If you mix notations, qs will merge the two items into an object:
        self.assertDictEqual(
            qs.parse('a[]=b&a[c]=d', parse_arrays=True),
            {'a': {'0': 'b', 'c': 'd'}})

        # You can also create arrays of objects:
        self.assertDictEqual(
            qs.parse('a[][b]=c', parse_arrays=True),
            {'a': {0: {'b': 'c'}}})

        # Some people use comma to join array, qs can parse it:
        self.assertDictEqual(
            qs.parse('a=b,c', comma=True),
            {'a': ['b', 'c']})

        # test double initialization of array with second index
        self.assertDictEqual(
            qs.parse('a[][1]=c', parse_arrays=True),
            {'a': {1: 'c'}})

        # test array notation on non array arguments
        self.assertDictEqual(
            qs.parse('a=c', parse_arrays=True),
            {'a': 'c'})

        # test over parameter limit in array notation
        self.assertDictEqual(
            qs.parse('a[0]=b&a[1]=c&b[]=x',
                     parameter_limit=1, parse_arrays=True),
            {'a': {0: 'b', 1: 'c'}})

    def test_basic_primitive_values(self):
        import src.qstion as qs
        # test some basic string values
        self.assertDictEqual(
            qs.parse('a=b'),
            {'a': 'b'})

        # with allowed primitive values, nothing should change
        self.assertDictEqual(
            qs.parse('a=b', parse_primitive=True),
            {'a': 'b'})

        # test with integer value
        self.assertDictEqual(
            qs.parse('a=1'),
            {'a': '1'})

        # with allowed primitive values, parsed should be integer
        self.assertDictEqual(
            qs.parse('a=1', parse_primitive=True),
            {'a': 1})

        # test with float value
        self.assertDictEqual(
            qs.parse('a=1.1'),
            {'a': '1.1'})

        # with allowed primitive values, parsed should be float
        self.assertDictEqual(
            qs.parse('a=1.1', parse_primitive=True),
            {'a': 1.1})

        # test with boolean value
        self.assertDictEqual(
            qs.parse('a=true&b=false'),
            {'a': 'true', 'b': 'false'})

        # with allowed primitive values, parsed should be boolean
        self.assertDictEqual(
            qs.parse('a=true&b=false', parse_primitive=True),
            {'a': True, 'b': False})

        # test with null value
        self.assertDictEqual(
            qs.parse('a=null&b=NULL&c=Null&d=None'),
            {'a': 'null', 'b': 'NULL', 'c': 'Null', 'd': 'None'})

        # with allowed primitive values, parsed should be None
        self.assertDictEqual(
            qs.parse('a=null&b=None', parse_primitive=True),
            {'a': None, 'b': None})

        # test array of primitive values
        self.assertDictEqual(
            qs.parse('a[]=1&a[]=2&a[]=3', parse_arrays=True),
            {'a': {0: '1', 1: '2', 2: '3'}})

        # with allowed primitive values, parsed should be array of integers
        self.assertDictEqual(
            qs.parse('a[]=1&a[]=2&a[]=3', parse_arrays=True,
                     parse_primitive=True),
            {'a': {0: 1, 1: 2, 2: 3}})

        # test classic array of primitive values
        self.assertDictEqual(
            qs.parse('a=[1,2,3]'),
            {'a': ['1', '2', '3']})

        # with allowed primitive values, parsed should be array of integers
        self.assertDictEqual(
            qs.parse('a=[1,2,3]', parse_primitive=True),
            {'a': [1, 2, 3]})

        # test classic array of primitive values with mixed types
        self.assertDictEqual(
            qs.parse('a=[1,a,true,Null]'),
            {'a': ['1', 'a', 'true', 'Null']})

        # with allowed primitive values, parsed should be array of primitives with mixed types
        self.assertDictEqual(
            qs.parse('a=[1,a,true,null]', parse_primitive=True),
            {'a': [1, 'a', True, None]})

        # test primitive values with non-strict keywords for bool and null
        self.assertDictEqual(
            qs.parse('a=True&b=False&c=none', parse_primitive=True),
            {'a': 'True', 'b': 'False', 'c': 'none'})

        # test primitive values with non-strict option
        self.assertDictEqual(
            qs.parse('a=True&b=False&c=NULL',
                     parse_primitive=True, primitive_strict=False),
            {'a': True, 'b': False, 'c': None})

    def test_parsing_from_dict(self):

        # test parsing from dictionary
        import src.qstion as qs
        self.assertDictEqual(
            qs.parse_from_dict({'a': 'b'}),
            {'a': 'b'})

        # test nested dictionary
        self.assertDictEqual(
            qs.parse_from_dict({'a[b]': 'c'}),
            {'a': {'b': 'c'}})


if __name__ == '__main__':
    unittest.main()



import typing as t
from html import unescape as unescape_html
import urllib.parse as up


class Unparsable(Exception):
    """Exception raised when query string cannot be parsed with QsParser"""
    pass


class ArrayLimitReached(Exception):
    """Exception raised when array limit is reached while parsing query string"""
    pass


class UnbalancedBrackets(Exception):
    """Exception raised when query string contains unbalanced brackets"""
    pass


class EmptyKey(Exception):
    """Exception raised when query string contains empty key and is not allowed"""
    pass


class QsNode:
    """
    Data structure to represent a query string as a tree for better manipulation of arrays and objects
    """
    key: str | int
    children: list['QsNode']
    value: t.Any

    def __init__(self, key: str | int, value: t.Any):
        self.key = key
        self.children = []
        if isinstance(value, type(self)):
            self.children = [value]
            self.value = None
        else:
            self.value = value
            self.children = []

    @classmethod
    def load(
        cls,
        parent_key: str | int,
        data: t.Any,
        filter: list = None
    ) -> 'QsNode':
        """
        Load a dictionary into a QsNode recursively

        Args:
            parent_key (str): parent key to be used as key for root node
            data (dict): dictionary to load
            filter (list): list of keys to filter - if key is not in filter, it will be ignored

        Returns:
            QsNode: current node
        """
        root = cls(parent_key, None)
        if filter and parent_key not in filter:
            return None
        if isinstance(data, dict):
            for k, v in data.items():
                child = cls.load(k, v, filter=filter)
                if child is not None:
                    root.children.append(child)
        elif isinstance(data, list):
            # list of dictionaries, indexed
            for i, v in enumerate(data):
                child = cls.load(i, v, filter=filter)
                if child is not None:
                    root.children.append(child)
        else:
            root.value = data
        return root

    def __getitem__(self, key: str):
        """
        Overload [] operator to get child by key
        """
        for child in self.children:
            if child.key == key:
                return child
        raise KeyError(key)

    def get(self, key: str, default: t.Any = None):
        """
        Method to get child by key with default value - behaves like dict.get
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key: str):
        """
        Overload in operator to check if child exists
        """
        for child in self.children:
            if child.key == key:
                return True
        return False

    def is_leaf(self):
        """
        Check if node is a leaf - has no children
        """
        return self.children is None or len(self.children) == 0

    def has_int_key(self):
        """
        Check if node has an integer key - used to determine whether node might be an array item
        """
        return isinstance(self.key, int)

    def is_array(self):
        """
        Check if node is an array - all children have integer keys - values might be nested
        """
        if self.value is None and all([child.has_int_key() for child in self.children]):
            return True

    def is_default_array(self):
        """
        Check if node is a default array - all children have integer keys and are leafs
        """
        if self.is_array() and all([child.is_leaf() for child in self.children]):
            return True

    def max_index(self) -> int:
        """
        Used to determine the next index for a new array item
        """
        # method is only called within _parse_array method but this is security measure
        return max([child.key for child in self.children]) if self.is_array() else -1

    def to_object_notation(self):
        """
        Transform self from possible array notation to default object notation - all integer indexes are replaced with 
        their string representation
        """
        if isinstance(self.key, int):
            self.key = str(self.key)
        for child in self.children:
            child.to_object_notation()

    def reorder(self):
        """
        Used to reorder children of a node - sorts them by key
        Only works if node is considered an array
        """
        if self.is_array():
            self.children.sort(key=lambda x: x.key)
            for child in self.children:
                child.reorder()

    def update(self, other: 'QsNode'):
        """
        Recursively update self with other node

        Args:
            other (QsNode): node to update self with

        Returns:
            None
        """
        if self.is_leaf() and other.is_leaf():
            self.merge_value(other)
        elif self.is_leaf() and not other.is_leaf():
            self.children = [QsNode(str(self.value), True)]
            self.value = None
            for child in other.children:
                if child.key in self:
                    self[child.key].update(child)
                else:
                    self.children.append(child)
            self.reorder()
        elif not self.is_leaf() and other.is_leaf():
            self.children.append(QsNode(other.value, True))
            self.reorder()
        else:
            for child in other.children:
                if child.key in self:
                    self[child.key].update(child)
                else:
                    self.children.append(child)
            self.reorder()

    def merge_value(self, other: 'QsNode'):
        """
        Merge values of two nodes

        Args:
            other (QsNode): node to merge value with

        Returns:
            None
        """
        if not isinstance(self.value, list):
            self.value = [self.value]
        if not isinstance(other.value, list):
            other.value = [other.value]
        self.value.extend(other.value)

    def serialize(self):
        """
        Represent self as a dictionary
        """
        if self.is_leaf():
            return self.value
        return {child.key: child.serialize() for child in self.children}

    def set_index(self, base: 'QsNode', array_limit: int = 20):
        """
        Set index of self and children recursively
        If base is provided, it is used to match children by key
        If array_limit is reached, ArrayLimitReached exception is raised

        Args:
            base (QsNode): base node to match children by key
            array_limit (int): array limit

        Returns:
            None
        """
        for child in self.children:
            match = base.get(child.key, None) if base else None
            child.set_index(match, array_limit=array_limit)
        for child in self.children:
            if child.key is None:
                child.key = base.max_index() + 1 if base else 0
            if child.has_int_key() and child.key > array_limit:
                raise ArrayLimitReached('Array limit reached')

    def is_empty(self):
        """
        Check if node is empty - has no value and no children
        """
        return self.is_leaf() and self.value is None


class QS:
    """
    Base class for query string parser and stringifier
    """
    _qs_tree: dict[str, QsNode]

    _depth: int = 5
    _depth: int = 5
    _parameter_limit: int = 1000
    _allow_dots: bool = False
    _array_limit: int = 20
    _parse_arrays: bool = False
    _allow_empty: bool = False
    _comma: bool = False

    def __init__(
            self,
            depth: int = 5,
            parameter_limit: int = 1000,
            allow_dots: bool = False,
            array_limit: int = 20,
            parse_arrays: bool = False,
            allow_empty: bool = False,
            comma: bool = False,
    ):
        """
        Args:
            args (list): list of arguments to parse
            depth (int): max depth of nested objects
            - e.g. depth=1 : a[b]=c -> {'a': {'b': 'c'}}
            - depth=1 : a[b][c]=d -> {'a': { '['b']['c']' : 'd'}}
            parameter_limit (int): max number of parameters to parse (in keyword count)
            allow_dots (bool): allow dot notation
            - e.g. a.b=c -> {'a': {'b': 'c'}}
            allow_sparse (bool): allow sparse arrays
            - e.g. a[1]=b&a[5]=c -> {'a': [,'b',,'c']}
            array_limit (int): max number of elements in array
            if limit is reached, array is converted to object
            parse_arrays (bool): parse array values as or keep object notation
            comma (bool): allow comma separated values
            - e.g. a=b,c -> {'a': ['b', 'c']}
        """
        self._qs_tree = {}
        self._max_depth = depth
        self._parameter_limit = parameter_limit
        self._allow_dots = allow_dots
        self._array_limit = array_limit
        self._parse_arrays = parse_arrays
        self._allow_empty = allow_empty
        self._comma = comma

    @staticmethod
    def _unq(arg: str, charset: str = 'utf-8', interpret_numeric_entities: bool = False) -> str:
        """
        Unquotes a string (removes url encoding).

        Args:
            arg (str): string to unquote
            charset (str): charset to use for unquoting
            interpret_numeric_entities (bool): interpret numeric entities in unicode

        Returns:
            str: unquoted string
        """
        try:
            arg_key, arg_val = arg.split('=')
        except ValueError:
            raise Unparsable('Unable to parse key')
        if interpret_numeric_entities:
            return (unescape_html(up.unquote(arg_key, charset)), unescape_html(up.unquote(arg_val, charset)))
        return (up.unquote(arg_key, charset), up.unquote(arg_val, charset))

    @staticmethod
    def _q(key: str, value: str, charset: str = 'utf-8') -> str:
        """
        Encodes a string (url encoding).

        Args:
            key (str): key to encode
            value (str): value to encode
            charset (str): charset to use for encoding

        Returns:
            str: encoded string
        """
        return f'{up.quote(key, encoding=charset, errors="xmlcharrefreplace")}={up.quote(value, encoding=charset, errors="xmlcharrefreplace")}'
